# Load the LARS library
library('lars')

# PART i
# Repeat the following steps at least 5 times. (You can just copy and paste 
# this entire section into R 5 times, but you should understand what each step
# does).
#
# First, we will generate an artificial dataset with 10 continuous covariates
# and a continuous response.  Our dataset will have 100 points in it.  All
# features will be selected independently at random from a normal distribution
# with variance 1.

x <- array(rnorm(100*10), dim=c(100,10))

# Now we generate the response variable.  It is a linear combination of two of
# the covariates: 3 and 7.  Feel free to change these.  It also has some 
# Gaussian noise added in.
noise <- rnorm(100)
y <- x[,3] + 10*x[,7] + 0.9*noise

# Try plotting y versus x[,3] with plot(x[,3],y) to see how difficult the
# relationship is to see.

# Now we examine the filtering score for each of the features.
cor(x,y)^2

# Can you explain what happens?  HINT: compare cor(x,y) to cor(x,x[,7])

# Now output the dataset for use later in part (e)
write.table(cbind(y,x),"featsel.dat",row.names=FALSE,col.names=FALSE)

# PART (b)
# Finally, we fit an L1-regularized linear regression model using LARS.
model <- lars(x,y)
plot.lars(model)

# PART (c) 
# Tweak all the numeric values below to your liking
x <- array(rnorm(150*50), dim=c(150,50))
noise <- rnorm(150)
y <- x[,2] + x[,3]+15*x[,45] + 1.2*noise
cor(x,y)^2
model <- lars(x,y)
plot.lars(model)


# PART (d)
# Repeat the following steps at least 5 times.

# First generate a dataset.  This time, we have 1000 variables and 1000
# data points.
x <- array(rnorm(1000*1000), dim=c(1000,1000))
err <- rnorm(1000)
y <- x[,3] + 10*x[,7] + err

# The filtering score can be computed quickly.
score <- cor(x,y)^2

# Get a list of features, sorted by filtering score
filtered <- sort(score, index.return=TRUE, decreasing=TRUE)[[2]]

# Look at the top two features; are they what you would expect?
filtered[1:2]

# LARS takes a while on a system this large (you can interrupt it before it's
# done by pressing the stop button in windows or with CTRL-C in unix)
#m <- lars(x,y)

# So try filtering all but the top 200 features. Now LARS runs quickly.
m <- lars(x[,filtered[1:200]],y)

# How did LARS do?  Note that now in order to interpret the indices reported
# by lars, you have to examine which indices they correspond to in the 
# filtered vector--for example, if lars says features 1 and 24 are the first
# two to enter the support, they are actually features filtered[1] and
# filtered[24].

plot.lars(m)


# PART (e)
# You've already output a dataset to featsel.dat above.  Good luck with YALE!