
## CS294: Practical Machine Learning (Fall 2006)
## PSet4, Part I: Hidden Markov Models (auxiliary functions)

# Reads text characters from file and processes word boundaries
readText <- function(filename, wordsPerSentence=0)
{
  # ensure consistent numeric encoding of letters
  alphabet <- strsplit("abcdefghijklmnopqrstuvwxyz_*", split="");
  alphabet <- alphabet[[1]];

  # read all words from file, ignoring whitespace
  words <- scan(file=filename, what="character");

  # separate words into individual characters
  chars <- strsplit(words, split="");

  # flatten character string, with underscore marking word boundaries
  if (wordsPerSentence > 0) {
    # break words into "sentences" of fixed length
    string <- list();
    ww = 1;
    ss = 1;
    sentence <- c();
    
    for (i in 1:length(chars)) {
      sentence <- c(sentence, chars[[i]], "_");

      if (ww == wordsPerSentence) {
        sentence <- sentence[1:(length(sentence)-1)];
        string[[ss]] <- factor(sentence, levels=alphabet);
        sentence <- c();
        ww <- 1;
        ss <- ss+1;
      }
      else {
        ww = ww+1;
      }
    }
  }
  else {
    # put all words in one long string
    string <- c();
    for (i in 1:length(chars)) {
      string <- c(string, chars[[i]], "_");
    }
    string <- string[1:(length(string)-1)];

    # encode as categorical data
    string <- factor(string, levels=alphabet);
  }

  return(string);
}

# Samples a random text sequence from a learned HMM
sampleText <- function(Nletter, model, levels)
{
  # randomly sample output sequence from HMM
  sequence <- sim.hmm(Nletter, model$tpm, model$Rho);

  # convert output to categorical form
  string <- cut(sequence, c(1:(length(levels)+1) - 0.5), levels);

  return(string);
}

# Perturb one text sequence into another via discrete, iid "noise"
perturbText <- function(string, noiseMatrix)
{
  # convert input string to numeric form
  sequence <- tidyup(string, levels(string));

  # randomly perturb each string element according to noiseMatrix
  numOutput <- nrow(noiseMatrix);
  noisy <- c();

  for (ww in 1:length(sequence)) {
    noisy[ww] <- sample(1:numOutput, 1, prob=noiseMatrix[,sequence[ww]]);
  }

  # convert output to categorical form
  noisy <- cut(noisy, c(1:(length(levels(string))+1) - 0.5), levels(string));

  return(noisy);
}

# Observation sequence log-likelihood (accessor for hmm.discnp)
loglike.hmm <- function(y, yval = NULL, tpm, Rho)
{
  y  <- tidyup(y, yval);
  nc <- length(y);
  fy <- ffun(y, Rho);

  if (length(tpm) > 1) {
    ispd <- revise.ispd(tpm);

    if (length(y) > 1) {
      rp <- recurse(fy, tpm, ispd, nc);
      ll <- sum(log(abs(rp$llc)));
    }
    else {
      ll <- log(fy %*% ispd);
    }
  }
  else {
    ll <- sum(log(fy));
  }

  return(ll);
}

# Marginal probabilities of hidden states (accessor for hmm.discnp)
margprob.hmm <- function(y, yval = NULL, tpm, Rho)
{
  y  <- tidyup(y, yval);
  nc <- length(y);
  fy <- ffun(y, Rho);

  if (length(tpm) > 1) {
    ispd  <- revise.ispd(tpm);
    rp    <- recurse(fy, tpm, ispd, nc);
    mprob <- rp$gamma;
  }
  else {
    mprob <- matrix(1,nrow=1,ncol=length(y));
  }
  
  return(mprob);
}



tidyup <- function(y,yval=NULL) {
  if(is.null(yval)) yval <- sort(unique(as.vector(y)))
  nval <- length(yval)
  if(is.matrix(y)) {
    nr <- nrow(y)
    nc <- ncol(y)
  }
  else {
    nr <- length(y)
    nc <- 1
  }
  matrix(match(y,yval),nr,nc)
  }
